﻿local Anzu = {}
local B = BardTools

Anzu.maxPlayers = 30

local L = {
["Eye of Anzu"] = GetSpellInfo(179202),
["Radiance of Anzu"] = GetSpellInfo(185239),
["Fel Bomb"] = GetSpellInfo(181753),
["Phantasmal Fel Bomb"] = GetSpellInfo(179219),
["Phantasmal Corruption"] = GetSpellInfo(181824),
["Phantasmal Cremation"] = GetSpellInfo(187344),
["Phantasmal Winds"] = GetSpellInfo(181956),
["Phantasmal Wounds"] = GetSpellInfo(182323),
["Dark Bindings"] = GetSpellInfo(185510),
["Fel Conduit"] = GetSpellInfo(181827),
["Fel Chakram"] = GetSpellInfo(182178),
["Phantasmal Obliteration"] = GetSpellInfo(179218),
["Shadow Escape"] = GetSpellInfo(181873),
["Spirit of Redemption"] = GetSpellInfo(20711),
["Purify"] = GetSpellInfo(527),
["Purify Spirit"] = GetSpellInfo(77130),
["Cleanse"] = GetSpellInfo(4987),
["Nature's Cure"] = GetSpellInfo(88423),
["Detox"] = GetSpellInfo(115450),
}

local N = {
["Corrupted Talonpriest1"] = 91543,
["Corrupted Talonpriest2"] = 93985,
["Phantasmal Resonance"] = 93625,
["Shadowfel Warden"] = 93968,
["Fel Raven"] = 93952,
["Shadow-Lord Iskar"] = 95067,
["Frostwall Rat"] = 80437,
}

local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitIsUnit = UnitIsUnit
local GetTime = GetTime
local UnitName = UnitName
local GetRaidRosterInfo = GetRaidRosterInfo
local GetNumGroupMembers = GetNumGroupMembers
local FullTargetName = B.FullTargetName
local GetInspectSpecialization = GetInspectSpecialization

local function CombatLogUnitName(unit)
    local name, realm = UnitName(unit)
    if (realm) then
        return name .. "-" .. realm
    else
        return name
    end
end

local function RaidUnitName(unit)
    -- Only intended for raid units
    -- Always returns player name and realm name, also when [unit] is from the same realm
    
    local name, realm = UnitName(unit)
    if (not realm) then
        realm = GetRealmName()
    end
    
    return name .. "-" .. realm
end

local oldUnitClass = UnitClass
local function UnitClass(unit)
    local class = select(2, oldUnitClass(unit))
    return class
end

local function NameToUnit(playerName)
    -- Only works for player raid units
    -- [playerName] must always be in the format "Playername-Servername", including when the player being referred to is from the same realm
    for i = 1,GetNumGroupMembers() do
        if (RaidUnitName("raid" .. i) == playerName) then
            return "raid" .. i
        end
    end
end

local function FelConduitInterrupter(unit)
    -- Returns true if [unit] is any class that has an interrupt of 15 seconds or less, and if [unit] is DPS
    local class = UnitClass(unit)
    local role = B:UnitRole(unit)
    if (role == "DAMAGER" and (class == "WARRIOR" or class == "DEATHKNIGHT" or class == "ROGUE" or class == "SHAMAN" or class == "PALADIN" or class == "MONK")) then
        return true
    end
end

local function SortByPriority(a, b)
    -- Sorts descending - {1,3,2} -> {3,2,1}
    if (a.priority > b.priority) then
        return true
    else
        return false
    end
end

local function GUIDToNPCID(GUID)
    local dash1 = strfind(GUID, "-")
    for i = 1,4 do
        dash1 = strfind(GUID, "-", dash1+1)
    end
    local dash2 = strfind(GUID, "-", dash1+1)
    
    return tonumber(strsub(GUID, dash1+1, dash2-1))
end

local function UnitNPCID(unit)
    local GUID = UnitGUID(unit)
    return GUIDToNPCID(GUID)
end

function Anzu:ShowHighlight(index)
    local f = _G["BardToolsAnzuFrame" .. index]
    
    if (Anzu.previousHighlight) then
        Anzu.previousHighlight:HideHighlight()
    end
    
    if (f.fadingIn) then
        f:StopFade()
    end
    
    f:ShowHighlight()
    Anzu.previousHighlight = f
end

function Anzu:HideHighlight()
    if (Anzu.previousHighlight) then
        Anzu.previousHighlight:HideHighlight()
    end
    Anzu.previousHighlight = nil
end

function B:UpdateAnzuFrameNames()
    if (BardToolsData.AnzuFrames.displayNames) then
        for i = 1,GetNumGroupMembers() do
            Anzu.frames[i].text:SetText(UnitName("raid" .. i))
        end
    else
        for i = 1,GetNumGroupMembers() do
            Anzu.frames[i].text:SetText()
        end
    end
end

function B:UpdateAnzuFrameSettings()
    if (InCombatLockdown()) then
        print("AnzuFrames: Can't change settings while in combat.")
        return
    end
    
    local anchorPoint
    local relativeAnchorPoint
    local horizontalGrowthMultiplier
    local verticalGrowthMultiplier
    
    if (BardToolsData.AnzuFrames.verticalGrowthDirection == "Up") then
        anchorPoint = "BOTTOM"
        verticalGrowthMultiplier = 1
    else
        anchorPoint = "TOP"
        verticalGrowthMultiplier = -1
    end
    
    if (BardToolsData.AnzuFrames.horizontalGrowthDirection == "Left") then
        anchorPoint = anchorPoint .. "RIGHT"
        horizontalGrowthMultiplier = -1
    else
        anchorPoint = anchorPoint .. "LEFT"
        horizontalGrowthMultiplier = 1
    end
    
    local function xFunc(i)
        if (BardToolsData.AnzuFrames.groupType == "Grid") then
            return (floor((i-1) / 5)) * (BardToolsData.AnzuFrames.width + BardToolsData.AnzuFrames.spacing) * horizontalGrowthMultiplier
        elseif (BardToolsData.AnzuFrames.groupType == "Vertical Stack") then
            return 0
        elseif (BardToolsData.AnzuFrames.groupType == "Horizontal Stack") then
            return (i-1) * (BardToolsData.AnzuFrames.width + BardToolsData.AnzuFrames.spacing) * horizontalGrowthMultiplier
        end
    end
    
    local function yFunc(i)
        if (BardToolsData.AnzuFrames.groupType == "Grid") then
            return ((i-1) % 5) * (BardToolsData.AnzuFrames.height + BardToolsData.AnzuFrames.spacing) * verticalGrowthMultiplier
        elseif (BardToolsData.AnzuFrames.groupType == "Vertical Stack") then
            return (i-1) * (BardToolsData.AnzuFrames.height + BardToolsData.AnzuFrames.spacing) * verticalGrowthMultiplier
        elseif (BardToolsData.AnzuFrames.groupType == "Horizontal Stack") then
            return 0
        end
    end
    
    for i = 1,Anzu.maxPlayers do
        local f = _G["BardToolsAnzuFrame" .. i]
        local width = BardToolsData.AnzuFrames.width
        local height = BardToolsData.AnzuFrames.height
        f:ClearAllPoints()
        f:SetPoint(anchorPoint, BardToolsAnzuFrameAnchor, "TOPLEFT", xFunc(i), yFunc(i))
        f:SetSize(width, height)
        
        if (width >= height) then
            f.icon:SetSize(height-7, height-7)
        else
            f.icon:SetSize(width-7, width-7)
        end
        
        if ((BardToolsConfig and BardToolsConfig:IsShown()) or (UnitExists("raid" .. i) and B.anzuHelperLoaded)) then
            if (not f:IsShown()) then
                f:Show()
            end
        else
            f:Hide()
        end
    end
end

function B:CreateAnzuFrames()
    if (BardToolsAnzuFrameAnchor) then
        return
    end
    
    local baseFrameTexture = {0.2, 0.2, 0.2}
    
    local a = CreateFrame("Frame", "BardToolsAnzuFrameAnchor", UIParent)
    a:SetMovable(true)
    local point, _, relativePoint, x, y = unpack(BardToolsData.AnzuFrames.anchorPoint)
    a:SetPoint(point, UIParent, relativePoint, x, y)
    a:SetSize(5,5)
    
    a.texture = a:CreateTexture()
    --a.texture:SetTexture(1,0,0)
    a.texture:SetAllPoints(a)
    
    Anzu.frames = {}
    
    for i = 1,Anzu.maxPlayers do
        local f = CreateFrame("Button", "BardToolsAnzuFrame" .. i, UIParent, "SecureActionButtonTemplate")
        f:RegisterForClicks("LeftButtonDown", "RightButtonDown")
        
        f:SetAttribute("type1", "macro")
        f:SetAttribute("type2", "macro")
        
        -- Change the below variable in-game to change the way the ability is cast on raid members (if the current one doesn't work)
        if (BardToolsData.AnzuFrames.macro == 2) then
            f:SetAttribute("macrotext", "/stopspelltarget" .. "\n" .. "/target raid" .. i .. "\n" .. "/click ExtraActionButton1" .. "\n" .. "/targetlasttarget")
        else
            f:SetAttribute("macrotext", "/stopspelltarget" .. "\n" .. "/use [@raid" .. i .. "] " .. (L["Eye of Anzu"] or "Noob"))
        end
        
        -- Set right click to dispel magic
        local spell = "Noob"
        local class = UnitClass("player")
        
        if (class == "PRIEST") then
            spell = L["Purify"]
        elseif (class == "SHAMAN") then
            spell = L["Purify Spirit"]
        elseif (class == "PALADIN") then
            spell = L["Cleanse"]
        elseif (class == "DRUID") then
            spell = L["Nature's Cure"]
        elseif (class == "MONK") then
            spell = L["Detox"]
        end
        
        f:SetAttribute("macrotext2", "/stopspelltarget" .. "\n" .. "/use [@raid" .. i .. "] " .. spell)
        
        function f:FadeIn(delay)
            if (f.fadingIn) then
                return
            else
                f.fadingIn = true
                f.fadeInStart = GetTime()
                f:SetScript("OnUpdate", 
                function()
                    local color = baseFrameTexture[1] + (1-baseFrameTexture[1])*((GetTime() - f.fadeInStart) / delay)
                    f.texture:SetTexture(color, color, color)
                    f.cooldown.texture:SetTexture(color, color, color)
                end)
                B:StartTimer(delay, function() f:StopFade() end, f:GetName() .. "StopFade")
                --B:StartTimer(delay, function() Anzu:ShowHighlight(i) end)
                B:StartTimer(delay, 
                    function() 
                        f.texture:SetTexture(baseFrameTexture[1], baseFrameTexture[2], baseFrameTexture[3])
                        f.cooldown.texture:SetTexture(baseFrameTexture[1], baseFrameTexture[2], baseFrameTexture[3])
                    end, f:GetName() .. "StopFadeRestoreTexture")
                f:StartCooldown(delay)
                f.cooldown:SetSwipeColor(0,0,0,0)
            end
        end
        
        function f:StopFade()
            f.fadingIn = false
            f.texture:SetTexture(baseFrameTexture[1], baseFrameTexture[2], baseFrameTexture[3])
            f.cooldown.texture:SetTexture(baseFrameTexture[1], baseFrameTexture[2], baseFrameTexture[3])
            f:SetScript("OnUpdate", nil)
            f:StopCooldown()
            B:StopTimer(f:GetName() .. "StopFade")
            B:StopTimer(f:GetName() .. "StopFadeRestoreTexture")
        end
        
        function f:StartCooldown(duration, color)
            if (not duration) then
                error("Need to specify a duration")
            end
            
            if (f.cooldown.active) then
                return
            else
                f.cooldown.active = true
                f.cooldown:SetCooldown(GetTime(), duration)
                f.cooldown:SetSwipeColor(0,0,0,0.6)
                f.cooldown.texture:Show()
                B:StartTimer(duration, function() f:StopCooldown() end, f:GetName() .. "Cooldown")
            end
        end
        
        function f:StopCooldown()
            f.cooldown:SetCooldown(0,0)
            f.cooldown.active = false
            f.cooldown.texture:Hide()
            B:StopTimer(f:GetName() .. "Cooldown")
        end
        
        function f:ShowHighlight()
            f.texture:SetTexture(1,1,1)
        end
        
        function f:HideHighlight()
            f.texture:SetTexture(unpack(baseFrameTexture))
        end
        
        f.old_SetSize = f.SetSize
        function f:SetSize(width, height)
            f:old_SetSize(width, height)
            f.texture:SetSize(width-3, height-3)
        end
        
        f.cooldown = CreateFrame("Cooldown", "BardToolsAnzuFrame" .. i .. "Cooldown", f)
        f.cooldown:SetAllPoints(f)
        f.cooldown:SetSwipeTexture(0)
        f.cooldown:SetSwipeColor(0,0,0,0.6)
        
        f.cooldown.texture = f:CreateTexture()
        f.cooldown.texture:SetPoint("CENTER", f, "CENTER", -1, -1)
        f.cooldown.texture:SetSize(15,15)
        f.cooldown.texture:Hide()
        f.cooldown.texture:SetDrawLayer("BORDER", 0)
        
        f.icon = f:CreateTexture(f:GetName() .. "Icon")
        f.icon:SetDrawLayer("BORDER", 1)
        f.icon:SetPoint("CENTER", f, "CENTER")
        
        f.texture = f:CreateTexture()
        f.texture:SetPoint("CENTER", f, "CENTER")
        f.texture:SetTexture(0.2,0.2,0.2)
        f.texture:SetDrawLayer("BACKGROUND", 1)
        
        f.mouseOverHighlight = f:CreateTexture()
        f.mouseOverHighlight:SetAllPoints(f)
        f.mouseOverHighlight:SetTexture(0.4,0.4,0.4)
        f.mouseOverHighlight:SetDrawLayer("BACKGROUND", 0)
        
        f.text = f:CreateFontString(nil, nil, "GameFontNormal")
        f.text:SetAllPoints(f)
        f.text:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
        f.text:SetTextColor(1,1,1)
        f.text:SetNonSpaceWrap(false)
        f.text:SetDrawLayer("BACKGROUND", 2)
        
        local function OnMouseDown()
            if (IsShiftKeyDown()) then
                BardToolsAnzuFrameAnchor:StartMoving()
                print("BardTools: You moved AnzuHelper frames by shift-dragging them.")
            end
        end
        
        local function OnMouseUp()
            BardToolsAnzuFrameAnchor:StopMovingOrSizing()
            BardToolsData.AnzuFrames.anchorPoint = {BardToolsAnzuFrameAnchor:GetPoint()}
        end
        
        f:SetScript("OnEnter", function() f.mouseOverHighlight:SetTexture(1,1,1) end)
        f:SetScript("OnLeave", function() f.mouseOverHighlight:SetTexture(0.4,0.4,0.4) end)
        f:SetScript("OnMouseDown", OnMouseDown)
        f:SetScript("OnMouseUp", function() BardToolsAnzuFrameAnchor:StopMovingOrSizing() end)
        
        tinsert(Anzu.frames, f)
    end
end

local I = {
["Fel Bomb"] = select(3, GetSpellInfo(181753)),
["Phantasmal Wounds"] = select(3, GetSpellInfo(182323)),
["Phantasmal Corruption"] = select(3, GetSpellInfo(187990)),
["Aspect of the Cheetah"] = select(3, GetSpellInfo(5118)),
}

function Anzu:SetIconTexture(unit, texture)
    local unitIndex = B:UnitIndex(unit)
    
    if (unit) then
        Anzu.frames[unitIndex].icon:SetTexture(texture)
    else
        Anzu.frames[unitIndex].icon:SetTexture()
    end
end

function Anzu:UpdateAuras(unit)
    if (strfind(unit, "raid") and strlen(unit) <= 6) then
        Anzu.units[unit].eye = UnitBuff(unit, L["Eye of Anzu"]) and true or false
        Anzu.units[unit].stacks = select(4, UnitDebuff(unit, L["Radiance of Anzu"]))
        Anzu.units[unit].felBomb = UnitDebuff(unit, L["Fel Bomb"])
        Anzu.units[unit].phantasmalFelBomb = UnitDebuff(unit, L["Phantasmal Fel Bomb"])
        Anzu.units[unit].phantasmalCorruption = UnitDebuff(unit, L["Phantasmal Corruption"])
        Anzu.units[unit].phantasmalWinds = UnitDebuff(unit, L["Phantasmal Winds"])
        Anzu.units[unit].phantasmalWounds = UnitDebuff(unit, L["Phantasmal Wounds"])
        Anzu.units[unit].darkBindings = UnitDebuff(unit, L["Dark Bindings"]) -- Mythic only
        Anzu.units[unit].felChakram = UnitDebuff(unit, L["Fel Chakram"])
        
        if (Anzu.units[unit].felBomb) then
            Anzu:SetIconTexture(unit, I["Fel Bomb"])
        elseif (Anzu.units[unit].phantasmalCorruption) then
            Anzu:SetIconTexture(unit, I["Phantasmal Corruption"])
        else
            Anzu:SetIconTexture(unit)
        end
        
        if (Anzu.units[unit].eye) then
            Anzu.eyeUnit = unit
        end
    elseif (unit == "player") then
        Anzu.playerHasEye = UnitBuff("player", L["Eye of Anzu"]) and true or false
    end
end

function Anzu:CombatLog(...)
    local event = select(2, ...)
    if (event == "SPELL_AURA_APPLIED") then
        local spellName = select(13, ...)
        local destGUID = select(8, ...)
        if (spellName == L["Phantasmal Winds"]) then
            Anzu.phantasmalWindsCount = Anzu.phantasmalWindsCount + 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Wounds"]) then
            Anzu.phantasmalWoundsCount = Anzu.phantasmalWoundsCount + 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Fel Bomb"]) then
            Anzu.phantasmalFelBombCount = Anzu.phantasmalFelBombCount + 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Corruption"]) then
            Anzu.phantasmalCorruption = true
            for i = 1,GetNumGroupMembers() do
                if (destGUID == UnitGUID("raid" .. i)) then
                    Anzu.phantasmalCorruptionUnit = "raid" .. i
                end
            end
            Anzu.update = true
        elseif (spellName == L["Eye of Anzu"]) then
            Anzu.update = true
        elseif (spellName == L["Spirit of Redemption"]) then -- AKA when a priest dies
            Anzu.update = true
        elseif (spellName == L["181873"]) then
            Anzu:ShadowEscape()
        end
        
    elseif (event == "SPELL_AURA_REMOVED") then
        local spellName = select(13, ...)
        if (spellName == L["Phantasmal Winds"]) then
            Anzu.phantasmalWindsCount = Anzu.phantasmalWindsCount - 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Wounds"]) then
            Anzu.phantasmalWoundsCount = Anzu.phantasmalWoundsCount - 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Fel Bomb"]) then
            Anzu.phantasmalFelBombCount = Anzu.phantasmalFelBombCount - 1
            Anzu.update = true
        elseif (spellName == L["Phantasmal Corruption"]) then
            Anzu.phantasmalCorruption = false
            Anzu.phantasmalCorruptionUnit = "none"
            Anzu.update = true
        end
        
    elseif (event == "SWING_DAMAGE" or event == "SWING_MISSED") then
        local sourceName = select(5, ...)
        if (sourceName == L["Fel Raven"]) then
            local destGUID = select(8, ...)
            Anzu.felRavenTargetGUID = destGUID
        end
        
    elseif (event == "SPELL_CAST_START") then
        local spellName = select(13, ...)
        if (spellName == L["Fel Conduit"]) then
            -- Starts a timer whose end is synchronized with the end of the entire Fel Conduit cast/channel sequence
            -- We will not track the buff that is applied during the channel because we aren't sure about certain specifics (spell name, etc)
            -- Timer is stopped anyway, when the channel/cast stops
            -- The cast time should be 1.9 sec
            -- The total channel time should be 9 sec
            Anzu.felConduit = true
            Anzu.update = true
            B:StartTimer(10.9, function() Anzu:FelConduitStop() end, "FelConduit")
        elseif (spellName == L["Phantasmal Obliteration"]) then
            -- Fade in dispel duty unit, if there is one and they are available
            -- Otherwise, fade in the healer with the lowest unit ID
            Anzu:FelBombCastStart()
            Anzu.phantasmalObliterationCasting = true
        end
        
    elseif (event == "SPELL_CAST_SUCCESS") then
        local spellName = select(13, ...)
        if (spellName == L["Phantasmal Obliteration"]) then
            Anzu.phantasmalObliterationCasting = false
        end
        
    elseif (event == "SPELL_INTERRUPT") then
        local extraSpellName = select(16, ...)
        if (extraSpellName == L["Fel Conduit"]) then
            Anzu:FelConduitStop()
        end
        
    elseif (event == "UNIT_DIED") then
        local destGUID, destName = select(8, ...)
        local unitNPCId = GUIDToNPCID(destGUID)
        if (unitNPCId == N["Corrupted Talonpriest1"] or unitNPCId == N["Corrupted Talonpriest2"] or destName == "Corrupted Talonpriest") then
            Anzu.corruptedTalonpriest = false
            if (Anzu.phantasmalObliterationCasting) then
                Anzu:FelBombCastStop()
            end
        elseif (unitNPCId == N["Shadowfel Warden"] or destName == "Shadowfel Warden") then
            Anzu.shadowfelWarden = false
            if (Anzu.felConduit) then
                Anzu:FelConduitStop()
            end
        elseif (unitNPCId == N["Fel Raven"] or destName == "Fel Raven") then
            Anzu.felRaven = false
        elseif (unitNPCId == N["Phantasmal Resonance"] or destName == "Phantasmal Resonance") then
            Anzu.phantasmalResonance = false
        elseif (unitNPCId == N["Shadow-Lord Iskar"] or destName == "Shadow-Lord Iskar") then
            B:UnloadAnzuModule()
        elseif (destGUID == UnitGUID(Anzu.eyeUnit)) then
            Anzu.eyeUnit = "none"
        end
        
        -- We need to update everything if anyone in the raid dies
        for i = 1,GetNumGroupMembers() do
            if (destGUID == UnitGUID("raid" .. i)) then
                Anzu.update = true
            end
        end
    end
end

function Anzu:Update()
    -- Here go ALL functions that sets highlights on frames
    -- It is okay to call other functions that set highlights from this function
    
    -- TODO list:
    
    -- 1:   Mythic specific mechanics:
    --      Dark Bindings
    --      Shadow Riposte
    --      Stacks don't reset, so we have to even them out
    
    if (Anzu.phantasmalCorruption) then
        Anzu:PhantasmalCorruption()
    elseif (Anzu.phantasmalFelBombCount > 0) then
        Anzu:FelBomb()
    elseif (Anzu.felConduit) then
        Anzu:FelConduit() 
    elseif (Anzu.phantasmalWindsCount > 0) then
        Anzu:SortPhantasmalWinds() 
    elseif (Anzu.phantasmalCorruption and Anzu:PhantasmalCremationStacks() <= 2) then
        Anzu:PhantasmalCorruption() 
    elseif (Anzu.phantasmalWoundsCount > 0) then
        Anzu:SortPhantasmalWounds()  
    elseif (Anzu.felRaven and Anzu.felRavenTargetGUID ~= "none") then
        Anzu:PhantasmalCorruptionPrevention()
    else
        Anzu:HideHighlight()
    end
end

function Anzu:MythicUpdate()
    -- Work in progress
end

function Anzu:ShadowEscape()
    Anzu.shadowEscapeCount = Anzu.shadowEscapeCount + 1
    if (Anzu.shadowEscapeCount >= 1) then
        Anzu.corruptedTalonpriest = true
    end
    if (Anzu.shadowEscapeCount >= 2) then
        Anzu.shadowfelWarden = true
    end
    if (Anzu.shadowEscapeCount >= 3) then
        Anzu.felRaven = true
    end
end

function Anzu:PhantasmalCremationStacks()
    -- Returns number of Phantasmal Cremation on the unit affected by Phantasmal Corruption
    local stacks = select(4, UnitDebuff(Anzu.phantasmalCorruptionUnit, L["Phantasmal Cremation"]))
    if (stacks) then
        return stacks
    else
        return 0
    end
end

function Anzu:PhantasmalCorruptionPrevention()
    if (Anzu.playerHasEye and UnitGUID("player") ~= Anzu.felRavenTargetGUID) then
        for i = 1,GetNumGroupMembers() do
            if (UnitGUID("raid" .. i) == Anzu.felRavenTargetGUID) then
                Anzu:ShowHighlight(i)
            end
        end
    else
        Anzu:HideHighlight()
    end
end

function Anzu:PhantasmalCorruption()
    if (Anzu.playerHasEye and Anzu.eyeUnit ~= Anzu.phantasmalCorruptionUnit) then
        if (Anzu.phantasmalCorruptionUnit == "none") then
            error("Anzu.phantasmalCorruptionUnit = 'none'.\nThis should definitely not happen.")
        else
            local unitIndex = B:UnitIndex(Anzu.phantasmalCorruptionUnit)
            Anzu:ShowHighlight(unitIndex)
        end
    end
end

function Anzu:FelConduitStop()
    -- Called when
    --  Fel Conduit is interrupted
    --  Fel Conduit finishes channeling (on a timer that starts upon casting)
    --  Shadowfel Warden dies while casting or channeling
    B:StopTimer("FelConduit")
    Anzu.felConduit = false
    Anzu.update = true
end

function Anzu:FelConduit()
    -- Fired when Shadowfel Warden starts casting Fel Conduit
    -- Highlights player on interrupt duty if there is one, but only if that player is available
    -- Otherwise, highlights a melee DPS
    -- This is probably the best reason to be able to highlight multiple units; preferably, all melee DPS should be highlighted
    -- MAJOR CAVEAT: If this guy is targeted by Chakram, stuff will fail. Unfortunately, there is no debuff to track Chakram targets
    
    if (Anzu.playerHasEye) then
        if (BardToolsData.interruptDuty) then
            -- If interrupter is assigned, give it to them if they are not dead, disconnected or in Spirit of Redemption form (they can't interrupt then)
            local unit = NameToUnit(BardToolsData.interruptDuty)
            if (not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                if (Anzu.eyeUnit == unit) then
                    Anzu:HideHighlight()
                else
                    Anzu:ShowHighlight(B:UnitIndex(unit))
                end
            else
                -- This is what to do if assigned interrupter is unavailable
                for i = 1,GetNumGroupMembers() do
                    local unit = "raid" .. i
                    if (not UnitDebuff(unit, L["Fel Chakram"]) and B:UnitRole(unit) == "TANK" and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit)) then
                        if (Anzu.eyeUnit == unit) then
                            Anzu:HideHighlight()
                        else
                            Anzu:ShowHighlight(i)
                        end
                        break
                    end
                end
            end
        else
            -- For now, just give it to any tank...
            for i = 1,GetNumGroupMembers() do
                local unit = "raid" .. i
                if (not UnitDebuff(unit, L["Fel Chakram"]) and B:UnitRole(unit) == "TANK" and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit)) then
                    if (Anzu.eyeUnit == unit) then
                        Anzu:HideHighlight()
                    else
                        Anzu:ShowHighlight(i)
                    end
                    break
                end
            end
        end
    else
        Anzu:HideHighlight()
    end
end

function Anzu:StopFadingAll()
    for k,v in pairs(Anzu.frames) do
        if (v.fadingIn) then
            v:StopFade()
        end
    end
end

function Anzu:FelBombCastStop()
    Anzu.phantasmalObliterationCasting = false
    Anzu:StopFadingAll()
end

function Anzu:FelBombCastStart()
    -- Fades in player on dispel duty if there is one, but only if that player is available
    -- Otherwise, fades in the healer with the lowest unit index
    
    if (Anzu.playerHasEye) then
        if (BardToolsData.dispelDuty) then
            -- If dispeller is assigned, give it to them if they are not dead, disconnected or in Spirit of Redemption form (they can't dispel then)
            local unit = NameToUnit(BardToolsData.dispelDuty)
            if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                local unitIndex = B:UnitIndex(unit)
                Anzu.frames[unitIndex]:FadeIn(4)
            else
                -- This is what to do if assigned dispeller is unavailable
                for i = 1,GetNumGroupMembers() do
                    local unit = "raid" .. i
                    if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                        local unitIndex = B:UnitIndex(unit)
                        Anzu.frames[unitIndex]:FadeIn(4)
                        break
                    end
                end
            end
        else
            -- For now, just give it to any healer...
            for i = 1,GetNumGroupMembers() do
                local unit = "raid" .. i
                if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                    local unitIndex = B:UnitIndex(unit)
                    Anzu.frames[unitIndex]:FadeIn(4)
                    break
                end
            end
        end
    else
        Anzu:HideHighlight()
    end
end

function Anzu:FelBomb()
    -- Highlights player on dispel duty if there is one, but only if that player is available
    -- Otherwise, highlights the healer with the lowest unit index
    
    if (Anzu.playerHasEye) then
        if (BardToolsData.dispelDuty) then
            -- If dispeller is assigned, give it to them if they are not dead, disconnected or in Spirit of Redemption form (they can't dispel then)
            local unit = NameToUnit(BardToolsData.dispelDuty)
            if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                if (Anzu.eyeUnit == unit) then
                    Anzu:HideHighlight()
                else
                    Anzu:ShowHighlight(B:UnitIndex(unit))
                end
            else
                -- This is what to do if assigned dispeller is unavailable
                for i = 1,GetNumGroupMembers() do
                    local unit = "raid" .. i
                    if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                        if (Anzu.eyeUnit == unit) then
                            Anzu:HideHighlight()
                        else
                            Anzu:ShowHighlight(i)
                        end
                        break
                    end
                end
            end
        else
            -- For now, just give it to any healer...
            for i = 1,GetNumGroupMembers() do
                local unit = "raid" .. i
                if (B:UnitRole(unit) == "HEALER" and not UnitDebuff(unit, L["Fel Chakram"]) and not UnitIsDeadOrGhost(unit) and UnitIsConnected(unit) and not UnitBuff(unit, L["Spirit of Redemption"])) then
                    if (Anzu.eyeUnit == unit) then
                        Anzu:HideHighlight()
                    else
                        Anzu:ShowHighlight(i)
                    end
                    break
                end
            end
        end
    else
        Anzu:HideHighlight()
    end
end

function Anzu:SortPhantasmalWounds()
    local priorityUnits = {}
    
    for i = 1,GetNumGroupMembers() do
        v = Anzu.units["raid" .. i]
        if (not UnitExists(v.unit)) then
            break -- Should never happen
        end
        
        if (v.phantasmalWounds) then
            v.priority = (GetNumGroupMembers()-B:UnitIndex(v.unit))/10000 -- Sort players of equal priority by unit index, where raid1 is 1 and raid40 is 40
        else
            v.priority = 0
        end
        
        if (v.priority > 0) then
            tinsert(priorityUnits, v)
        end
    end
    
    -- Sort descending:
    sort(priorityUnits, SortByPriority)
    
    -- This is the check that will be used if player is NOT one of the units with Phantasmal Winds,
    -- which will be the case after player just had Phantasmal Winds removed by Eye:
    if (Anzu.playerHasEye and priorityUnits[1]) then
        Anzu:ShowHighlight(B:UnitIndex(priorityUnits[1].unit))
        return
    end
    
    -- This is the check that will be used if player is one of the units with Phantasmal Winds:
    -- Highlights whichever unit is just below the player on the priority list, or nothing if player is last
    -- TODO: If player is last: Will it be a hassle to highlight the unit that should have the orb after all Phantasmal Winds are gone?
    for k,v in pairs(priorityUnits) do
        if (UnitIsUnit(v.unit, "player")) then
            if (priorityUnits[k+1]) then
                Anzu:ShowHighlight(B:UnitIndex(priorityUnits[k+1].unit))
                return
            end
        end
    end
    
    -- Hide any highlights if both of the above checks fail:
    Anzu:HideHighlight()
end

function Anzu:SortPhantasmalWinds()
    local priorityUnits = {}
    
    for i = 1,GetNumGroupMembers() do
        v = Anzu.units["raid" .. i]
        if (not UnitExists(v.unit)) then
            break -- Should never happen, but oh well.
        end
        
        local role = B:UnitRole(v.unit)
        local class = UnitClass(v.unit)
        
        if (v.phantasmalWinds) then
            v.priority = (GetNumGroupMembers()-B:UnitIndex(v.unit))/10000 -- Sort players of equal priority by unit index, where raid1 is 1 and raid40 is 40
            
            -- Immobile Healers > Immobile DPS > Tanks > Healers > DPS
            -- The whole process should take less than 3 seconds, so the order doesn't matter much anyway
            if (role == "HEALER" and (class == "SHAMAN" or class == "PALADIN" or class == "MONK")) then
                v.priority = v.priority + 5
            elseif (role == "DAMAGER" and (class == "MAGE" or class == "WARLOCK" or class == "PRIEST" or class == "DRUID" or class == "SHAMAN")) then
                v.priority = v.priority + 4
            elseif (role == "TANK") then
                v.priority = v.priority + 3
            elseif (role == "HEALER") then
                v.priority = v.priority + 2
            else
                v.priority = v.priority + 1
            end
            
        else
            v.priority = 0
        end
        
        if (v.priority > 0) then
            tinsert(priorityUnits, v)
        end
    end
    
    -- Sort descending:
    sort(priorityUnits, SortByPriority)
    
    -- This is the check that will be used if player is NOT one of the units with Phantasmal Winds,
    -- which will be the case after player just had Phantasmal Winds removed by eye:
    if (Anzu.playerHasEye and priorityUnits[1]) then
        Anzu:ShowHighlight(B:UnitIndex(priorityUnits[1].unit))
        return
    end
    
    -- This is the check that will be used if player is one of the units with Phantasmal Winds:
    -- Highlights whichever unit is just below the player on the priority list, or nothing if player is last
    -- TODO: If player is last: Will it be a hassle to highlight the unit that should have the orb after all Phantasmal Winds are gone?
    for k,v in pairs(priorityUnits) do
        if (UnitIsUnit(v.unit, "player")) then
            if (priorityUnits[k+1]) then
                Anzu:ShowHighlight(B:UnitIndex(priorityUnits[k+1].unit))
                return
            end
        end
    end
    
    -- Hide any highlights if both of the above checks fail:
    Anzu:HideHighlight()
end

function Anzu:Reset()
    Anzu.update = true
    Anzu.felConduit = false
    Anzu.phantasmalObliterationCasting = false
    Anzu.phantasmalCorruption = false
    Anzu.phantasmalCorruptionUnit = "none"
    Anzu.phantasmalFelBombCount = 0
    Anzu.phantasmalWindsCount = 0
    Anzu.phantasmalWoundsCount = 0
    Anzu.shadowEscapeCount = 0
    Anzu.felRavenTargetGUID = "none"
    
    Anzu.eyeUnit = "none"
    Anzu.playerHasEye = false
    
    Anzu.corruptedTalonpriest = false
    Anzu.shadowfelWarden = false
    Anzu.felRaven = false
    Anzu.phantasmalResonance = false
    
    Anzu:HideHighlight()
    Anzu:StopFadingAll()
    
    B:StopAllTimers()
    
    local groupType = "raid"
    
    Anzu.units = {}
    for i = 1,Anzu.maxPlayers do
        Anzu.units[groupType .. i] = {
            unit = groupType .. i,
            eye = false,
            stacks = 0,
            priority = 0,
            darkBindings = false,
            felBomb = false,
            phantasmalFelBomb = false,
            phantasmalCorruption = false,
            phantasmalWinds = false,
            phantasmalWounds = false,
        }
    end
end

function Anzu:GroupRosterUpdate()
    B:UpdateAnzuFrameSettings()
    if (BardToolsData.AnzuFrames.displayNames) then
        B:UpdateAnzuFrameNames()
    end
end

local function OnEvent(self, event, ...)
    if (event == "UNIT_AURA") then
        Anzu:UpdateAuras(...)
    elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
        Anzu:CombatLog(...)
    elseif (event == "GROUP_ROSTER_UPDATE" or event == "GROUP_JOINED") then
        Anzu:GroupRosterUpdate()
    elseif (event == "PLAYER_REGEN_ENABLED") then
        B:UnloadAnzuModule()
    end
end

local function OnUpdate()
    if (Anzu.update) then
        Anzu.update = false
        Anzu:Update()
    end
end

function B:LoadAnzuModule()
    B.anzuHelperLoaded = true
    print("BardTools: AnzuHelper loaded.")
    
    local f = BardToolsAnzuEventListener or CreateFrame("Frame", "BardToolsAnzuEventListener", UIParent)
    f:SetScript("OnEvent", OnEvent)
    f:SetScript("OnUpdate", OnUpdate)
    f:RegisterEvent("UNIT_AURA")
    f:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    f:RegisterEvent("GROUP_ROSTER_UPDATE")
    f:RegisterEvent("GROUP_JOINED")
    
    B:CreateAnzuFrames()
    B:UpdateAnzuFrameSettings()
    B:UpdateAnzuFrameNames()
    Anzu:Reset()
end

function B:UnloadAnzuModule()
    local f = BardToolsAnzuEventListener
    
    if (InCombatLockdown()) then
        f:RegisterEvent("PLAYER_REGEN_ENABLED")
        return
    end
        
    B.anzuHelperLoaded = false
    print("BardTools: AnzuHelper stopped.")
    
    Anzu:Reset()
    
    f:UnregisterAllEvents()
    f:SetScript("OnEvent", nil)
    f:SetScript("OnUpdate", nil)
    
    B:UpdateAnzuFrameSettings()
end



